﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace XamDataChart
{
    using System;
    using System.Globalization;
    using System.Runtime.Serialization;

    /// <summary>
    /// Define the period of the time.
    /// </summary>
    [DataContract]
    public struct Period : IComparable<Period>, IComparable, IEquatable<Period>
    {
        /// <summary>
        /// The undefined period.
        /// </summary>
        public static readonly Period Undefined = new Period(0, TimeUnit.Year);

        /// <summary>
        /// The step to the seconds.
        /// </summary>
        private static readonly int[] Step = { 1, 60, 3600, 86400, 604800, 2592000, 7776000, 31104000 };

        /// <summary>
        /// Initializes a new instance of the <see cref="Period"/> struct.
        /// </summary>
        /// <param name="period">The period string.</param>
        public Period(string period)
            : this()
        {
            var temp = Parse(period);
            this.Duration = temp.Duration;
            this.Unit = temp.Unit;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Period"/> struct.
        /// </summary>
        /// <param name="duration">The duration value.</param>
        /// <param name="unit">The unit of the time.</param>
        public Period(double duration, TimeUnit unit)
            : this()
        {
            this.Duration = duration;
            this.Unit = unit;
        }

        /// <summary>
        /// Gets or sets the unit of the time.
        /// </summary>
        [DataMember]
        public TimeUnit Unit { get; set; }

        /// <summary>
        /// Gets or sets the duration of the period.
        /// </summary>
        [DataMember]
        public double Duration { get; set; }

        /// <summary>
        /// Parse the input period.
        /// </summary>
        /// <param name="period">
        /// The period string.
        /// </param>
        /// <returns>
        /// The <see cref="Period"/> struct.
        /// </returns>
        /// <exception cref="InvalidCastException">
        /// If the postfix for unit is invalid to parse, it will throw the exception.
        /// </exception>
        public static Period Parse(string period)
        {
            period = period.ToUpper().Trim();

            var result = new Period();

            if (period.EndsWith(Const.Year, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Year);
                result.Unit = TimeUnit.Year;
            }
            else if (period.EndsWith(Const.Quarter, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Quarter);
                result.Unit = TimeUnit.Quarter;
            }
            else if (period.EndsWith(Const.Month, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Month);
                result.Unit = TimeUnit.Month;
            }
            else if (period.EndsWith(Const.Week, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Week);
                result.Unit = TimeUnit.Week;
            }
            else if (period.EndsWith(Const.Day, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Day);
                result.Unit = TimeUnit.Day;
            }
            else if (period.EndsWith(Const.Hour, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Hour);
                result.Unit = TimeUnit.Hour;
            }
            else if (period.EndsWith(Const.Minute, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Minute);
                result.Unit = TimeUnit.Minute;
            }
            else if (period.EndsWith(Const.Second, StringComparison.OrdinalIgnoreCase))
            {
                result.Duration = ParseDuration(period, Const.Second);
                result.Unit = TimeUnit.Second;
            }
            else
            {
                throw new InvalidCastException(string.Format("The input string, {0}, cannot be converted to period.", period));
            }

            return result;
        }

        /// <summary>
        /// Try to parse the period string to object.
        /// </summary>
        /// <param name="period">The period string.</param>
        /// <param name="result">The result period object.</param>
        /// <returns>
        /// If true, it means the period is converted correct. If the input string is invalid, it returns false.
        /// </returns>
        public static bool TryParse(string period, out Period result)
        {
            try
            {
                result = Parse(period);
                return true;
            }
            catch (InvalidCastException)
            {
            }

            result = Undefined;
            return false;
        }

        /// <summary>
        /// Provide the equal comparison of the two period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="p2">The other period object.</param>
        /// <returns>True, if the two periods are equal. Otherwise, it returns false.</returns>
        public static bool operator ==(Period p1, Period p2)
        {
            return p1.Equals(p2);
        }

        /// <summary>
        /// Provide the not equal comparison of the two period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="p2">The other period object.</param>
        /// <returns>False, if the two periods are equal. Otherwise, it returns true.</returns>
        public static bool operator !=(Period p1, Period p2)
        {
            return !(p1 == p2);
        }

        /// <summary>
        /// Override the plus operation of the period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="p2">The other period object.</param>
        /// <returns>The total period. The unit of the result is same to the left term.</returns>
        public static Period operator +(Period p1, Period p2)
        {
            var temp = p2.Convert(p1.Unit);
            var duration = p1.Duration + temp.Duration;

            return new Period(duration, p1.Unit);
        }

        /// <summary>
        /// Override the plus operation of the period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="addition">The addition value of the period duration.</param>
        /// <returns>The total period. The unit of the result is same to the left term.</returns>
        public static Period operator +(Period p1, double addition)
        {
            return new Period(p1.Duration + addition, p1.Unit);
        }

        /// <summary>
        /// Override the subtract operation of the period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="p2">The other period object.</param>
        /// <returns>The total period. The unit of the result is same to the left term.</returns>
        public static Period operator -(Period p1, Period p2)
        {
            var temp = p2.Convert(p1.Unit);
            var duration = p1.Duration - temp.Duration;

            return new Period(duration, p1.Unit);
        }

        /// <summary>
        /// Override the plus operation of the period.
        /// </summary>
        /// <param name="p1">The period object.</param>
        /// <param name="diff">The difference of the period duration.</param>
        /// <returns>The total period. The unit of the result is same to the left term.</returns>
        public static Period operator -(Period p1, double diff)
        {
            return new Period(p1.Duration - diff, p1.Unit);
        }

        /// <summary>
        /// Convert the period among the different units.
        /// </summary>
        /// <param name="newUnit">The new unit.</param>
        /// <returns>
        /// The <see cref="Period"/> of the new unit.
        /// </returns>
        public Period Convert(TimeUnit newUnit)
        {
            var diff = (double)Step[(int)this.Unit] / Step[(int)newUnit];
            var duration = this.Duration * diff;

            return new Period(duration, newUnit);
        }

        /// <summary>
        /// Compare the period with the other.
        /// </summary>
        /// <param name="other">The other.</param>
        /// <returns>
        /// <para>
        /// A value that indicates the relative order of the objects being compared. <br />
        /// If the value is less than zero, this instance precedes to <c>other</c> in the sort order.<br />
        /// If the value is zero, this instance occurs the same position in the sort order as <c>other</c>. <br />
        /// If the value is greater than zero, this instance follows <c>other</c> in the sort order.
        /// </para>
        /// </returns>
        public int CompareTo(Period other)
        {
            var period = other.Convert(this.Unit);
            return Math.Sign(this.Duration - period.Duration);
        }

        /// <summary>
        /// Compare the period with the other.
        /// </summary>
        /// <param name="obj">The object.</param>
        /// <returns>
        /// <para>
        /// A value that indicates the relative order of the objects being compared. <br />
        /// If the value is less than zero, this instance precedes to <c>obj</c> in the sort order.<br />
        /// If the value is zero, this instance occurs the same position in the sort order as <c>obj</c>. <br />
        /// If the value is greater than zero, this instance follows <c>obj</c> in the sort order.
        /// </para>
        /// </returns>
        /// <exception cref="InvalidCastException">The compared object is not valid period.</exception>
        public int CompareTo(object obj)
        {
            if (obj is Period)
            {
                return this.CompareTo((Period)obj);
            }

            throw new InvalidCastException("The input object is not a valid Period.");
        }

        /// <summary>
        /// Output a string representing the period value.
        /// </summary>
        /// <returns>
        /// The <see cref="string"/> of the period.
        /// </returns>
        public override string ToString()
        {
            switch (this.Unit)
            {
                case TimeUnit.Second:
                    return this.Duration + Const.Second;
                case TimeUnit.Minute:
                    return this.Duration + Const.Minute;
                case TimeUnit.Hour:
                    return this.Duration + Const.Hour;
                case TimeUnit.Day:
                    return this.Duration + Const.Day;
                case TimeUnit.Week:
                    return this.Duration + Const.Week;
                case TimeUnit.Month:
                    return this.Duration + Const.Month;
                case TimeUnit.Quarter:
                    return this.Duration + Const.Quarter;
                case TimeUnit.Year:
                    return this.Duration + Const.Year;
                default:
                    return this.Duration.ToString();
            }
        }

        /// <summary>
        /// Check whether the two periods are equal.
        /// </summary>
        /// <param name="other">The other period.</param>
        /// <returns>
        /// True, if the two periods are equal. Otherwise, it returns false.
        /// </returns>
        public bool Equals(Period other)
        {
            return this.CompareTo(other) == 0;
        }

        /// <summary>
        /// Check whether the two periods are equal.
        /// </summary>
        /// <param name="obj">The other object.</param>
        /// <returns>
        /// True, if the two periods are equal. Otherwise, it returns false. If the object is null, then it returns false.
        /// </returns>
        public override bool Equals(object obj)
        {
            if (ReferenceEquals(null, obj))
            {
                return false;
            }

            return obj is Period && this.Equals((Period)obj);
        }

        /// <summary>
        /// Get the hash code of the object.
        /// </summary>
        /// <returns>
        /// The <see cref="int"/> of hash code.
        /// </returns>
        public override int GetHashCode()
        {
            unchecked
            {
                return ((int)this.Unit * 397) ^ this.Duration.GetHashCode();
            }
        }

        /// <summary>
        /// Parse the duration in the input period string.
        /// </summary>
        /// <param name="period">The input period string.</param>
        /// <param name="unit">The unit string.</param>
        /// <returns>
        /// The <see cref="double"/> value of the period.
        /// </returns>
        /// <exception cref="InvalidCastException">
        /// If the period number is invalid to parse, it will throw the exception.
        /// </exception>
        private static double ParseDuration(string period, string unit)
        {
            var duration = period.Substring(0, period.Length - unit.Length).Trim().ToUpper();
            double result;

            if (double.TryParse(duration, NumberStyles.Any, NumberFormatInfo.InvariantInfo, out result))
            {
                return result;
            }

            throw new InvalidCastException(string.Format("The input string, {0}, cannot be converted to period.", period));
        }
    }
}
